/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { IChannelAttributes, IFluidDataStoreRuntime, IChannelStorageService } from "@fluidframework/datastore-definitions/internal";
import { ISequencedDocumentMessage } from "@fluidframework/driver-definitions/internal";
import { ISummaryTreeWithStats } from "@fluidframework/runtime-definitions/internal";
import { IFluidSerializer, SharedObject } from "@fluidframework/shared-object-base/internal";
import { ITaskManager, ITaskManagerEvents } from "./interfaces.js";
/**
 * {@inheritDoc ITaskManager}
 *
 * @sealed
 * @legacy
 * @alpha
 */
export declare class TaskManagerClass extends SharedObject<ITaskManagerEvents> implements ITaskManager {
    /**
     * Mapping of taskId to a queue of clientIds that are waiting on the task.  Maintains the consensus state of the
     * queue, even if we know we've submitted an op that should eventually modify the queue.
     */
    private readonly taskQueues;
    private readonly opWatcher;
    private readonly queueWatcher;
    private readonly abandonWatcher;
    private readonly connectionWatcher;
    private readonly completedWatcher;
    private messageId;
    /**
     * Tracks the most recent pending op for a given task
     */
    private readonly latestPendingOps;
    /**
     * Tracks tasks that are this client is currently subscribed to.
     */
    private readonly subscribedTasks;
    /**
     * Map to track tasks that have pending complete ops.
     */
    private readonly pendingCompletedTasks;
    /**
     * Returns the clientId. Will return a placeholder if the runtime is detached and not yet assigned a clientId.
     */
    private get clientId();
    /**
     * Returns a ReadOnlyInfo object to determine current read/write permissions.
     */
    private get readOnlyInfo();
    /**
     * Constructs a new task manager. If the object is non-local an id and service interfaces will
     * be provided
     *
     * @param runtime - data store runtime the task queue belongs to
     * @param id - optional name of the task queue
     */
    constructor(id: string, runtime: IFluidDataStoreRuntime, attributes: IChannelAttributes);
    private submitVolunteerOp;
    private submitAbandonOp;
    private submitCompleteOp;
    /**
     * {@inheritDoc ITaskManager.volunteerForTask}
     */
    volunteerForTask(taskId: string): Promise<boolean>;
    /**
     * {@inheritDoc ITaskManager.subscribeToTask}
     */
    subscribeToTask(taskId: string): void;
    /**
     * {@inheritDoc ITaskManager.abandon}
     */
    abandon(taskId: string): void;
    /**
     * {@inheritDoc ITaskManager.assigned}
     */
    assigned(taskId: string): boolean;
    /**
     * {@inheritDoc ITaskManager.queued}
     */
    queued(taskId: string): boolean;
    /**
     * {@inheritDoc ITaskManager.subscribed}
     */
    subscribed(taskId: string): boolean;
    /**
     * {@inheritDoc ITaskManager.complete}
     */
    complete(taskId: string): void;
    /**
     * {@inheritDoc ITaskManager.canVolunteer}
     */
    canVolunteer(): boolean;
    /**
     * Create a summary for the task manager
     *
     * @returns the summary of the current state of the task manager
     */
    protected summarizeCore(serializer: IFluidSerializer): ISummaryTreeWithStats;
    /**
     * {@inheritDoc @fluidframework/shared-object-base#SharedObject.loadCore}
     */
    protected loadCore(storage: IChannelStorageService): Promise<void>;
    /***/
    protected initializeLocalCore(): void;
    /**
     * {@inheritDoc @fluidframework/shared-object-base#SharedObject.onDisconnect}
     */
    protected onDisconnect(): void;
    /**
     * {@inheritDoc @fluidframework/shared-object-base#SharedObject.onConnect}
     */
    protected onConnect(): void;
    /**
     * Override resubmit core to avoid resubmission on reconnect.  On disconnect we accept our removal from the
     * queues, and leave it up to the user to decide whether they want to attempt to re-enter a queue on reconnect.
     */
    protected reSubmitCore(): void;
    /**
     * Process a task manager operation
     *
     * @param message - the message to prepare
     * @param local - whether the message was sent by the local client
     * @param localOpMetadata - For local client messages, this is the metadata that was submitted with the message.
     * For messages from a remote client, this will be undefined.
     */
    protected processCore(message: ISequencedDocumentMessage, local: boolean, localOpMetadata: unknown): void;
    private addClientToQueue;
    private removeClientFromQueue;
    private removeClientFromAllQueues;
    /**
     * Will replace all instances of the placeholderClientId with the current clientId. This should only be called when
     * transitioning from detached to attached and this.runtime.clientId is defined.
     */
    private replacePlaceholderInAllQueues;
    private scrubClientsNotInQuorum;
    protected applyStashedOp(content: any): void;
}
//# sourceMappingURL=taskManager.d.ts.map